# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import os
import platform
import socket
import sys
import time
import warnings
from collections import defaultdict

import mock
import pytest
from pytest_localserver.http import ContentServer
from pytest_localserver.https import DEFAULT_CERTIFICATE

import elasticapm
from elasticapm.base import Client
from elasticapm.conf import _in_fips_mode
from elasticapm.conf.constants import ERROR

try:
    from elasticapm.utils.simplejson_encoder import dumps as simplejson_dumps
except ImportError:
    simplejson_dumps = None
from tests.fixtures import DummyTransport, TempStoreClient
from tests.utils import assert_any_record_contains


@pytest.mark.parametrize("elasticapm_client", [{"environment": "production"}], indirect=True)
def test_service_info(elasticapm_client):
    service_info = elasticapm_client.get_service_info()
    assert service_info["name"] == elasticapm_client.config.service_name
    assert service_info["environment"] == elasticapm_client.config.environment == "production"
    assert service_info["language"] == {"name": "python", "version": platform.python_version()}
    assert service_info["agent"]["name"] == "python"
    assert service_info["agent"]["activation_method"] == "unknown"


@pytest.mark.parametrize(
    "elasticapm_client", [{"environment": "production", "service_node_name": "my_node"}], indirect=True
)
def test_service_info_node_name(elasticapm_client):
    service_info = elasticapm_client.get_service_info()
    assert service_info["name"] == elasticapm_client.config.service_name
    assert service_info["environment"] == elasticapm_client.config.environment == "production"
    assert service_info["language"] == {"name": "python", "version": platform.python_version()}
    assert service_info["agent"]["name"] == "python"
    assert service_info["node"]["configured_name"] == "my_node"


def test_process_info(elasticapm_client):
    process_info = elasticapm_client.get_process_info()
    assert process_info["pid"] == os.getpid()
    assert process_info["ppid"] == os.getppid()
    assert "argv" not in process_info
    elasticapm_client.config.update("1", include_process_args=True)
    with mock.patch.object(sys, "argv", ["a", "b", "c"]):
        process_info = elasticapm_client.get_process_info()
    assert process_info["argv"] == ["a", "b", "c"]


def test_system_info(elasticapm_client):
    # mock docker/kubernetes data here to get consistent behavior if test is run in docker
    with mock.patch("elasticapm.utils.cgroup.get_cgroup_container_metadata") as mocked:
        mocked.return_value = {}
        system_info = elasticapm_client.get_system_info()
    assert {"detected_hostname", "architecture", "platform"} == set(system_info.keys())
    assert system_info["detected_hostname"] == elasticapm.utils.getfqdn()


@pytest.mark.parametrize("elasticapm_client", [{"hostname": "my_custom_hostname"}], indirect=True)
def test_system_info_hostname_configurable(elasticapm_client):
    # mock docker/kubernetes data here to get consistent behavior if test is run in docker
    system_info = elasticapm_client.get_system_info()
    assert system_info["configured_hostname"] == "my_custom_hostname"


@pytest.mark.parametrize("elasticapm_client", [{"global_labels": "az=us-east-1,az.rack=8"}], indirect=True)
def test_global_labels(elasticapm_client):
    data = elasticapm_client.build_metadata()
    assert data["labels"] == {"az": "us-east-1", "az_rack": "8"}


def test_docker_kubernetes_system_info(elasticapm_client):
    # mock docker/kubernetes data here to get consistent behavior if test is run in docker
    with mock.patch("elasticapm.utils.cgroup.get_cgroup_container_metadata") as mock_metadata:
        mock_metadata.return_value = {"container": {"id": "123"}, "kubernetes": {"pod": {"uid": "456"}}}
        system_info = elasticapm_client.get_system_info()
    assert system_info["container"] == {"id": "123"}
    assert system_info["kubernetes"] == {"pod": {"uid": "456", "name": elasticapm.utils.getfqdn().split(".")[0]}}


@mock.patch.dict(
    "os.environ",
    {
        "KUBERNETES_NODE_NAME": "node",
        "KUBERNETES_NAMESPACE": "namespace",
        "KUBERNETES_POD_NAME": "pod",
        "KUBERNETES_POD_UID": "podid",
    },
)
def test_docker_kubernetes_system_info_from_environ():
    # initialize agent only after overriding environment
    elasticapm_client = TempStoreClient(metrics_interval="0ms")
    # mock docker/kubernetes data here to get consistent behavior if test is run in docker
    with mock.patch("elasticapm.utils.cgroup.get_cgroup_container_metadata") as mock_metadata:
        mock_metadata.return_value = {}
        system_info = elasticapm_client.get_system_info()
    assert "kubernetes" in system_info
    assert system_info["kubernetes"] == {
        "pod": {"uid": "podid", "name": "pod"},
        "node": {"name": "node"},
        "namespace": "namespace",
    }


@mock.patch.dict(
    "os.environ",
    {
        "KUBERNETES_NODE_NAME": "node",
        "KUBERNETES_NAMESPACE": "namespace",
        "KUBERNETES_POD_NAME": "pod",
        "KUBERNETES_POD_UID": "podid",
    },
)
def test_docker_kubernetes_system_info_from_environ_overrides_cgroups():
    # initialize agent only after overriding environment
    elasticapm_client = TempStoreClient(metrics_interval="0ms")
    # mock docker/kubernetes data here to get consistent behavior if test is run in docker
    with mock.patch("elasticapm.utils.cgroup.get_cgroup_container_metadata") as mock_metadata, mock.patch(
        "socket.gethostname"
    ) as mock_gethostname:
        mock_metadata.return_value = {"container": {"id": "123"}, "kubernetes": {"pod": {"uid": "456"}}}
        mock_gethostname.return_value = "foo"
        system_info = elasticapm_client.get_system_info()
    assert "kubernetes" in system_info

    assert system_info["kubernetes"] == {
        "pod": {"uid": "podid", "name": "pod"},
        "node": {"name": "node"},
        "namespace": "namespace",
    }
    assert system_info["container"] == {"id": "123"}


@mock.patch.dict("os.environ", {"KUBERNETES_NAMESPACE": "namespace"})
def test_docker_kubernetes_system_info_except_hostname_from_environ():
    # initialize agent only after overriding environment
    elasticapm_client = TempStoreClient(metrics_interval="0ms")
    # mock docker/kubernetes data here to get consistent behavior if test is run in docker
    with mock.patch("elasticapm.utils.cgroup.get_cgroup_container_metadata") as mock_metadata, mock.patch(
        "socket.gethostname"
    ) as mock_gethostname:
        mock_metadata.return_value = {}
        mock_gethostname.return_value = "foo"
        system_info = elasticapm_client.get_system_info()
    assert "kubernetes" in system_info
    assert system_info["kubernetes"] == {
        "pod": {"name": elasticapm.utils.getfqdn().split(".")[0]},
        "namespace": "namespace",
    }


def test_config_by_environment():
    with mock.patch.dict("os.environ", {"ELASTIC_APM_SERVICE_NAME": "envapp", "ELASTIC_APM_SECRET_TOKEN": "envtoken"}):
        client = TempStoreClient(metrics_interval="0ms")
        assert client.config.service_name == "envapp"
        assert client.config.secret_token == "envtoken"
        assert client.config.disable_send is False
    with mock.patch.dict("os.environ", {"ELASTIC_APM_DISABLE_SEND": "true"}):
        client = TempStoreClient(metrics_interval="0ms")
        assert client.config.disable_send is True
    client.close()


def test_config_non_string_types():
    """
    tests if we can handle non string types as configuration, e.g.
    Value types from django-configuration
    """

    class MyValue(object):
        def __init__(self, content):
            self.content = content

        def __str__(self):
            return str(self.content)

        def __repr__(self):
            return repr(self.content)

    client = TempStoreClient(
        server_url="localhost", service_name=MyValue("bar"), secret_token=MyValue("bay"), metrics_interval="0ms"
    )
    assert isinstance(client.config.secret_token, str)
    assert isinstance(client.config.service_name, str)
    client.close()


@pytest.mark.parametrize("elasticapm_client", [{"transport_class": "tests.fixtures.DummyTransport"}], indirect=True)
def test_custom_transport(elasticapm_client):
    assert isinstance(elasticapm_client._transport, DummyTransport)


@pytest.mark.skipif(simplejson_dumps is None, reason="no test without simplejson")
@pytest.mark.parametrize(
    "elasticapm_client", [{"transport_json_serializer": "elasticapm.utils.simplejson_encoder.dumps"}], indirect=True
)
def test_custom_transport_json_serializer(elasticapm_client):
    assert elasticapm_client._transport._json_serializer == simplejson_dumps


@pytest.mark.parametrize("elasticapm_client", [{"processors": []}], indirect=True)
def test_empty_processor_list(elasticapm_client):
    assert elasticapm_client.processors == []


@pytest.mark.skipif(platform.system() == "Windows", reason="Flaky test on windows")
@pytest.mark.parametrize(
    "sending_elasticapm_client",
    [{"transport_class": "elasticapm.transport.http.Transport", "async_mode": False}],
    indirect=True,
)
@pytest.mark.parametrize("validating_httpserver", [{"app": ContentServer}], indirect=True)
@mock.patch("elasticapm.transport.base.TransportState.should_try")
def test_send_remote_failover_sync(should_try, sending_elasticapm_client, caplog):
    sending_elasticapm_client.httpserver.code = 400
    sending_elasticapm_client.httpserver.content = "go away"
    should_try.return_value = True

    # test error
    with caplog.at_level("ERROR", "elasticapm.transport"):
        sending_elasticapm_client.capture_message("foo", handled=False)
    sending_elasticapm_client._transport._flushed.wait(timeout=1)
    assert sending_elasticapm_client._transport.state.did_fail()
    assert_any_record_contains(caplog.records, "go away")

    # test recovery
    sending_elasticapm_client.httpserver.code = 202
    sending_elasticapm_client.capture_message("bar", handled=False)
    sending_elasticapm_client.close()
    assert not sending_elasticapm_client._transport.state.did_fail()


@mock.patch("elasticapm.transport.base.TransportState.should_try", return_value=True)
def test_send_remote_failover_sync_non_transport_exception_error(should_try, caplog):
    client = Client(
        server_url="http://example.com",
        service_name="app_name",
        secret_token="secret",
        transport_class="tests.fixtures.MockSendHTTPTransport",
        metrics_interval="0ms",
        metrics_sets=[],
    )

    # test error
    client._transport.send_mock.side_effect = ValueError("oopsie")
    with caplog.at_level("ERROR", "elasticapm.transport"):
        client.capture_message("foo", handled=False)
    try:
        client._transport.flush()
    except ValueError:
        # give flush a bit more room because we may take a bit more than the max timeout to flush
        client._transport._flushed.wait(timeout=1)
    assert client._transport.state.did_fail()
    assert_any_record_contains(caplog.records, "oopsie", "elasticapm.transport")

    # test recovery
    client._transport.send_mock.side_effect = None
    client.capture_message("foo", handled=False)
    try:
        client._transport.flush()
    except ValueError:
        # give flush a bit more room because we may take a bit more than the max timeout to flush
        client._transport._flushed.wait(timeout=1)
    # We have a race here with the queue where we would end up checking for did_fail before the message
    # is being handled by the queue, so sleep a bit and retry to give it enough time
    retries = 0
    while client._transport.state.did_fail() and retries < 3:
        time.sleep(0.1)
        retries += 1
    assert not client._transport.state.did_fail()


@pytest.mark.parametrize("validating_httpserver", [{"skip_validate": True}], indirect=True)
def test_send(sending_elasticapm_client):
    sending_elasticapm_client.queue("x", {})
    sending_elasticapm_client.close()
    request = sending_elasticapm_client.httpserver.requests[0]
    expected_headers = {
        "Content-Type": "application/x-ndjson",
        "Content-Encoding": "gzip",
        "Authorization": "Bearer %s" % sending_elasticapm_client.config.secret_token,
        "User-Agent": "apm-agent-python/%s (myapp)" % elasticapm.VERSION,
    }
    seen_headers = dict(request.headers)
    for k, v in expected_headers.items():
        assert seen_headers[k] == v

    # Commented out per @beniwohli
    # TODO: figure out why payload size is larger than 400 on windows / 2.7
    # assert 250 < request.content_length < 400


@pytest.mark.flaky(reruns=3)  # test is flaky on Windows
@pytest.mark.parametrize("sending_elasticapm_client", [{"disable_send": True}], indirect=True)
def test_send_not_enabled(sending_elasticapm_client):
    assert sending_elasticapm_client.config.disable_send
    sending_elasticapm_client.queue("x", {})
    sending_elasticapm_client.close()

    assert len(sending_elasticapm_client.httpserver.requests) == 0


@pytest.mark.parametrize(
    "sending_elasticapm_client",
    [{"transport_class": "elasticapm.transport.http.Transport", "async_mode": False}],
    indirect=True,
)
def test_client_shutdown_sync(sending_elasticapm_client):
    sending_elasticapm_client.capture_message("x")
    sending_elasticapm_client.close()
    assert len(sending_elasticapm_client.httpserver.requests) == 1


def test_call_end_twice(elasticapm_client):
    elasticapm_client.begin_transaction("celery")

    elasticapm_client.end_transaction("test-transaction", 200)
    elasticapm_client.end_transaction("test-transaction", 200)


@pytest.mark.skipif(_in_fips_mode() is True, reason="cannot disable verify_server_cert in fips mode")
@pytest.mark.parametrize("elasticapm_client", [{"verify_server_cert": False}], indirect=True)
def test_client_disables_ssl_verification(elasticapm_client):
    assert not elasticapm_client.config.verify_server_cert
    assert not elasticapm_client._transport._verify_server_cert


@pytest.mark.parametrize("sending_elasticapm_client", [{"server_cert": DEFAULT_CERTIFICATE}], indirect=True)
def test_server_cert_pinning(sending_elasticapm_client):
    assert sending_elasticapm_client._transport._server_cert == DEFAULT_CERTIFICATE


@pytest.mark.parametrize(
    "setting,url,result",
    [
        ("", "/foo/bar", False),
        ("*", "/foo/bar", True),
        ("/foo/bar,/baz", "/foo/bar", True),
        ("/foo/bar,/baz", "/baz", True),
        ("/foo/bar,/bazz", "/baz", False),
        ("/foo/*/bar,/bazz", "/foo/bar", False),
        ("/foo/*/bar,/bazz", "/foo/ooo/bar", True),
        ("*/foo/*/bar,/bazz", "/foo/ooo/bar", True),
        ("*/foo/*/bar,/bazz", "/baz/foo/ooo/bar", True),
    ],
)
def test_should_ignore_url(elasticapm_client, setting, url, result):
    elasticapm_client.config.update(1, transaction_ignore_urls=setting)
    assert elasticapm_client.should_ignore_url(url) is result


@pytest.mark.parametrize("sending_elasticapm_client", [{"disable_send": True}], indirect=True)
def test_disable_send(sending_elasticapm_client):
    assert sending_elasticapm_client.config.disable_send

    sending_elasticapm_client.capture("Message", message="test", data={"logger": "test"})

    assert len(sending_elasticapm_client.httpserver.requests) == 0


@pytest.mark.parametrize("elasticapm_client", [{"service_name": "@%&!"}], indirect=True)
def test_invalid_service_name_disables_send(elasticapm_client):
    assert len(elasticapm_client.config.errors) == 1
    assert "SERVICE_NAME" in elasticapm_client.config.errors

    assert elasticapm_client.config.disable_send


def test_empty_transport_disables_send():
    client = TempStoreClient(service_name="x", transport_class=None, metrics_interval="0ms")
    assert len(client.config.errors) == 1
    assert "TRANSPORT_CLASS" in client.config.errors

    assert client.config.disable_send
    client.close()


@pytest.mark.parametrize("sending_elasticapm_client", [{"service_name": "*"}], indirect=True)
@mock.patch("elasticapm.transport.base.Transport.send")
def test_config_error_stops_error_send(mock_send, sending_elasticapm_client):
    assert sending_elasticapm_client.config.disable_send is True
    sending_elasticapm_client.capture_message("bla", handled=False)
    assert mock_send.call_count == 0


@pytest.mark.parametrize("sending_elasticapm_client", [{"service_name": "*"}], indirect=True)
@mock.patch("elasticapm.transport.base.Transport.send")
def test_config_error_stops_transaction_send(mock_send, sending_elasticapm_client):
    assert sending_elasticapm_client.config.disable_send is True
    sending_elasticapm_client.begin_transaction("test")
    sending_elasticapm_client.end_transaction("test", "OK")
    sending_elasticapm_client.close()
    assert mock_send.call_count == 0


@pytest.mark.parametrize(
    "elasticapm_client,expected",
    [
        ({"server_url": "http://localhost"}, "http://localhost/intake/v2/events"),
        ({"server_url": "http://localhost/"}, "http://localhost/intake/v2/events"),
        ({"server_url": "http://localhost:8200"}, "http://localhost:8200/intake/v2/events"),
        ({"server_url": "http://localhost:8200/"}, "http://localhost:8200/intake/v2/events"),
        ({"server_url": "http://localhost/a"}, "http://localhost/a/intake/v2/events"),
        ({"server_url": "http://localhost/a/"}, "http://localhost/a/intake/v2/events"),
        ({"server_url": "http://localhost:8200/a"}, "http://localhost:8200/a/intake/v2/events"),
        ({"server_url": "http://localhost:8200/a/"}, "http://localhost:8200/a/intake/v2/events"),
    ],
    indirect=["elasticapm_client"],
)
def test_server_url_joining(elasticapm_client, expected):
    assert elasticapm_client._api_endpoint_url == expected


@pytest.mark.parametrize(
    "version",
    [
        ("2", "7", "0"),
        ("3", "3", "0"),
        ("3", "4", "0"),
        ("3", "5", "0"),
    ],
)
@mock.patch("platform.python_version_tuple")
def test_python_version_deprecation(mock_python_version_tuple, version):
    warnings.simplefilter("always")

    client_config = {}
    client_config["central_config"] = "false"
    client_config["cloud_provider"] = False
    client_config["transport_class"] = "tests.fixtures.DummyTransport"
    client_config["metrics_interval"] = "0ms"

    mock_python_version_tuple.return_value = version
    e = None
    with pytest.warns(DeprecationWarning, match="agent only supports"):
        try:
            e = elasticapm.Client(**client_config)
        finally:
            if e:
                e.close()


def test_recording(elasticapm_client):
    assert elasticapm_client.capture_message("x") is not None
    try:
        1 / 0
    except ZeroDivisionError:
        assert elasticapm_client.capture_exception() is not None
    assert elasticapm_client.begin_transaction("test") is not None
    with elasticapm.capture_span("x") as x_span:
        assert x_span is not None
    assert elasticapm_client.end_transaction("ok", "ok") is not None

    elasticapm_client.config.update("1", recording=False)
    assert not elasticapm_client.config.is_recording
    assert elasticapm_client.capture_message("x") is None
    try:
        1 / 0
    except ZeroDivisionError:
        assert elasticapm_client.capture_exception() is None
    assert elasticapm_client.begin_transaction("test") is None
    with elasticapm.capture_span("x") as x_span:
        assert x_span is None
    assert elasticapm_client.end_transaction("ok", "ok") is None


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        {"enabled": True, "metrics_interval": "30s", "central_config": "true"},
        {"enabled": False, "metrics_interval": "30s", "central_config": "true"},
    ],
    indirect=True,
)
def test_client_enabled(elasticapm_client):
    if elasticapm_client.config.enabled:
        assert elasticapm_client.config.is_recording
        for manager in elasticapm_client._thread_managers.values():
            assert manager.is_started()
    else:
        assert not elasticapm_client.config.is_recording
        for manager in elasticapm_client._thread_managers.values():
            assert not manager.is_started()


def test_excepthook(elasticapm_client):
    try:
        raise Exception("hi!")
    except Exception:
        type_, value, traceback = sys.exc_info()
        elasticapm_client._excepthook(type_, value, traceback)

    assert elasticapm_client.events[ERROR]


def test_check_server_version(elasticapm_client):
    assert elasticapm_client.server_version is None
    assert elasticapm_client.check_server_version(gte=(100, 5, 10))
    assert elasticapm_client.check_server_version(lte=(100, 5, 10))

    elasticapm_client.server_version = (7, 15, 0)
    assert elasticapm_client.check_server_version(gte=(7,))
    assert not elasticapm_client.check_server_version(gte=(8,))
    assert not elasticapm_client.check_server_version(lte=(7,))
    assert elasticapm_client.check_server_version(lte=(8,))
    assert elasticapm_client.check_server_version(gte=(7, 12), lte=(7, 15))
    assert elasticapm_client.check_server_version(gte=(7, 15), lte=(7, 15))
    assert elasticapm_client.check_server_version(gte=(7, 15), lte=(7, 16))
    assert not elasticapm_client.check_server_version(gte=(7, 12), lte=(7, 13))
    assert not elasticapm_client.check_server_version(gte=(7, 16), lte=(7, 18))


@pytest.mark.parametrize(
    "elasticapm_client,expected",
    [
        ({"service_version": "v2"}, " v2"),
        ({"service_version": "v2 \x00"}, " v2 _"),
        ({}, ""),
    ],
    indirect=["elasticapm_client"],
)
def test_user_agent(elasticapm_client, expected):
    assert elasticapm_client.get_user_agent() == "apm-agent-python/{} (myapp{})".format(elasticapm.VERSION, expected)


def test_label_without_client():
    elasticapm.label(foo="foo")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#  BSD 3-Clause License
#
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#
#
#

import mock

from elasticapm.conf.constants import ERROR


def test_chained_cause_exception(elasticapm_client):
    try:
        try:
            1 / 0
        except ZeroDivisionError as zde:
            raise ValueError("bla") from zde
    except ValueError:
        elasticapm_client.capture_exception()
    error = elasticapm_client.events[ERROR][0]
    assert error["exception"]["type"] == "ValueError"
    assert error["exception"]["cause"][0]["type"] == "ZeroDivisionError"


def test_chained_context_exception(elasticapm_client):
    try:
        try:
            1 / 0
        except ZeroDivisionError:
            int("zero")
    except ValueError:
        elasticapm_client.capture_exception()
    error = elasticapm_client.events[ERROR][0]
    assert error["exception"]["type"] == "ValueError"
    assert error["exception"]["cause"][0]["type"] == "ZeroDivisionError"


def test_chained_context_exception_suppressed(elasticapm_client):
    try:
        try:
            1 / 0
        except ZeroDivisionError:
            raise ValueError("bla") from None
    except ValueError:
        elasticapm_client.capture_exception()
    error = elasticapm_client.events[ERROR][0]
    assert error["exception"]["type"] == "ValueError"
    assert "cause" not in error["exception"]


def test_chained_context_exception_max(elasticapm_client):
    with mock.patch("elasticapm.events.EXCEPTION_CHAIN_MAX_DEPTH", 1):
        try:
            try:
                1 / 0
            except ZeroDivisionError:
                try:
                    1 + "a"
                except TypeError as e:
                    raise ValueError("bla") from e
        except ValueError:
            elasticapm_client.capture_exception()
        error = elasticapm_client.events[ERROR][0]
        assert error["exception"]["type"] == "ValueError"
        assert error["exception"]["cause"][0]["type"] == "TypeError"
        assert "cause" not in error["exception"]["cause"][0]



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import urllib.parse

import pytest
import urllib3
from urllib3._version import __version__ as urllib3_version

from elasticapm.conf import constants
from elasticapm.conf.constants import SPAN, TRANSACTION
from elasticapm.traces import capture_span
from elasticapm.utils.disttracing import TraceParent


def test_urllib3(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction")
    expected_sig = "GET {0}".format(parsed_url.netloc)
    with capture_span("test_name", "test_type"):
        pool = urllib3.PoolManager(timeout=0.1)

        url = "http://{0}/hello_world".format(parsed_url.netloc)
        r = pool.request("GET", url)

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    expected_signatures = {"test_name", expected_sig}

    assert {t["name"] for t in spans} == expected_signatures

    assert len(spans) == 2

    assert spans[0]["name"] == expected_sig
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["context"]["http"]["url"] == url
    assert spans[0]["context"]["http"]["status_code"] == 200
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["context"]["service"]["target"]["type"] == "http"
    assert spans[0]["context"]["service"]["target"]["name"] == f"127.0.0.1:{parsed_url.port}"
    assert spans[0]["parent_id"] == spans[1]["id"]
    assert spans[0]["outcome"] == "success"

    assert spans[1]["name"] == "test_name"
    assert spans[1]["type"] == "test_type"
    assert spans[1]["parent_id"] == transactions[0]["id"]


@pytest.mark.parametrize("status_code", [400, 500])
def test_urllib3_error(instrument, elasticapm_client, waiting_httpserver, status_code):
    waiting_httpserver.serve_content("", code=status_code)
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    elasticapm_client.begin_transaction("transaction")
    expected_sig = "GET {0}".format(parsed_url.netloc)
    with capture_span("test_name", "test_type"):
        pool = urllib3.PoolManager(timeout=0.1)

        url = "http://{0}/hello_world".format(parsed_url.netloc)
        r = pool.request("GET", url)

    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    assert spans[0]["name"] == expected_sig
    assert spans[0]["type"] == "external"
    assert spans[0]["subtype"] == "http"
    assert spans[0]["context"]["http"]["url"] == url
    assert spans[0]["context"]["http"]["status_code"] == status_code
    assert spans[0]["context"]["destination"]["service"] == {
        "name": "",
        "resource": "127.0.0.1:%d" % parsed_url.port,
        "type": "",
    }
    assert spans[0]["parent_id"] == spans[1]["id"]
    assert spans[0]["outcome"] == "failure"


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
def test_trace_parent_propagation_sampled(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    elasticapm_client.begin_transaction("transaction")
    pool = urllib3.PoolManager(timeout=0.1)
    r = pool.request("GET", url)
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transactions[0]["trace_id"]
    assert trace_parent.span_id == spans[0]["id"]
    assert trace_parent.trace_options.recorded
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict

    if elasticapm_client.config.use_elastic_traceparent_header:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME in headers
        assert headers[constants.TRACEPARENT_HEADER_NAME] == headers[constants.TRACEPARENT_LEGACY_HEADER_NAME]
    else:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME not in headers


@pytest.mark.parametrize(
    "elasticapm_client",
    [
        pytest.param({"use_elastic_traceparent_header": True}, id="use_elastic_traceparent_header-True"),
        pytest.param({"use_elastic_traceparent_header": False}, id="use_elastic_traceparent_header-False"),
    ],
    indirect=True,
)
def test_trace_parent_propagation_unsampled(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction")
    transaction_object.is_sampled = False
    pool = urllib3.PoolManager(timeout=0.1)
    r = pool.request("GET", url)
    elasticapm_client.end_transaction("MyView")
    spans = elasticapm_client.events[SPAN]

    assert not spans

    headers = waiting_httpserver.requests[0].headers
    assert constants.TRACEPARENT_HEADER_NAME in headers
    trace_parent = TraceParent.from_string(
        headers[constants.TRACEPARENT_HEADER_NAME], tracestate_string=headers[constants.TRACESTATE_HEADER_NAME]
    )
    assert trace_parent.trace_id == transaction_object.trace_parent.trace_id
    assert trace_parent.span_id == transaction_object.id
    assert not trace_parent.trace_options.recorded
    # Check that sample_rate was correctly placed in the tracestate
    assert constants.TRACESTATE.SAMPLE_RATE in trace_parent.tracestate_dict
    if elasticapm_client.config.use_elastic_traceparent_header:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME in headers
        assert headers[constants.TRACEPARENT_HEADER_NAME] == headers[constants.TRACEPARENT_LEGACY_HEADER_NAME]
    else:
        assert constants.TRACEPARENT_LEGACY_HEADER_NAME not in headers


@pytest.mark.parametrize(
    "is_sampled", [pytest.param(True, id="is_sampled-True"), pytest.param(False, id="is_sampled-False")]
)
def test_tracestate_propagation(instrument, elasticapm_client, waiting_httpserver, is_sampled):
    traceparent = TraceParent.from_string(
        "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03", "foo=bar,baz=bazzinga"
    )

    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction", trace_parent=traceparent)
    transaction_object.is_sampled = is_sampled
    pool = urllib3.PoolManager(timeout=0.1)
    r = pool.request("GET", url)
    elasticapm_client.end_transaction("MyView")
    headers = waiting_httpserver.requests[0].headers
    assert headers[constants.TRACESTATE_HEADER_NAME] == "foo=bar,baz=bazzinga"


@pytest.mark.parametrize("elasticapm_client", [{"transaction_max_spans": 1}], indirect=True)
def test_span_only_dropped(instrument, elasticapm_client, waiting_httpserver):
    """test that urllib3 instrumentation does not fail if no parent span can be found"""
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    transaction_object = elasticapm_client.begin_transaction("transaction")
    for i in range(2):
        with capture_span("test", "test"):
            pool = urllib3.PoolManager(timeout=0.1)
            pool.request("GET", url)
    elasticapm_client.end_transaction("bla", "OK")
    trace_parent_1 = TraceParent.from_string(waiting_httpserver.requests[0].headers[constants.TRACEPARENT_HEADER_NAME])
    trace_parent_2 = TraceParent.from_string(waiting_httpserver.requests[1].headers[constants.TRACEPARENT_HEADER_NAME])

    assert trace_parent_1.span_id != transaction_object.id
    # second request should use transaction id as span id because there is no span
    assert trace_parent_2.span_id == transaction_object.id


def test_url_sanitization(instrument, elasticapm_client, waiting_httpserver):
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    url = url.replace("http://", "http://user:pass@")
    transaction_object = elasticapm_client.begin_transaction("transaction")
    pool = urllib3.PoolManager(timeout=0.1)
    r = pool.request("GET", url)
    elasticapm_client.end_transaction("MyView")
    transactions = elasticapm_client.events[TRANSACTION]
    span = elasticapm_client.spans_for_transaction(transactions[0])[0]

    assert "pass" not in span["context"]["http"]["url"]


@pytest.mark.parametrize(
    "is_sampled", [pytest.param(True, id="is_sampled-True"), pytest.param(False, id="is_sampled-False")]
)
@pytest.mark.parametrize(
    "instance_headers",
    [pytest.param(True, id="instance-headers-set"), pytest.param(False, id="instance-headers-not-set")],
)
@pytest.mark.parametrize(
    "header_arg,header_kwarg",
    [
        pytest.param(True, False, id="args-set"),
        pytest.param(False, True, id="kwargs-set"),
        pytest.param(False, False, id="both-not-set"),
    ],
)
def test_instance_headers_are_respected(
    instrument, elasticapm_client, waiting_httpserver, is_sampled, instance_headers, header_arg, header_kwarg
):
    traceparent = TraceParent.from_string(
        "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03", "foo=bar,baz=bazzinga"
    )

    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    transaction_object = elasticapm_client.begin_transaction("transaction", trace_parent=traceparent)
    transaction_object.is_sampled = is_sampled
    pool = urllib3.HTTPConnectionPool(
        parsed_url.hostname,
        parsed_url.port,
        maxsize=1,
        block=True,
        headers={"instance": "true"} if instance_headers else None,
    )
    if header_arg:
        if urllib3_version.startswith("2"):
            args = ("GET", url, None, None, {"args": "true"})
        else:
            args = ("GET", url, None, {"args": "true"})
    else:
        args = ("GET", url)
    if header_kwarg:
        kwargs = {"headers": {"kwargs": "true"}}
    else:
        kwargs = {}
    r = pool.request(*args, **kwargs)
    request_headers = waiting_httpserver.requests[0].headers
    # all combinations should have the "traceparent" header
    assert "traceparent" in request_headers, (instance_headers, header_arg, header_kwarg)
    if header_arg:
        assert "args" in request_headers
    if header_kwarg:
        assert "kwargs" in request_headers
    if instance_headers and not (header_arg or header_kwarg):
        assert "instance" in request_headers


def test_connection_pool_urlopen_does_not_crash_with_many_args(instrument, elasticapm_client, waiting_httpserver):
    """Mimics ConnectionPool.urlopen error path with broken connection, see #1928"""
    waiting_httpserver.serve_content("")
    url = waiting_httpserver.url + "/hello_world"
    parsed_url = urllib.parse.urlparse(url)
    pool = urllib3.HTTPConnectionPool(
        parsed_url.hostname,
        parsed_url.port,
        maxsize=1,
        block=True,
    )
    retry = urllib3.util.Retry(10)
    elasticapm_client.begin_transaction("transaction")
    r = pool.urlopen("GET", url, None, {"args": "true"}, retry, False, False)
    assert r.status == 200



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

from elasticapm.utils.wsgi import get_environ, get_headers, get_host


def test_get_headers_tuple_as_key():
    result = dict(get_headers({("a", "tuple"): "foo"}))
    assert result == {}


def test_get_headers_coerces_http_name():
    result = dict(get_headers({"HTTP_ACCEPT": "text/plain"}))
    assert "accept" in result
    assert result["accept"] == "text/plain"


def test_get_headers_coerces_content_type():
    result = dict(get_headers({"CONTENT_TYPE": "text/plain"}))
    assert "content-type" in result
    assert result["content-type"] == "text/plain"


def test_get_headers_coerces_content_length():
    result = dict(get_headers({"CONTENT_LENGTH": "134"}))
    assert "content-length" in result
    assert result["content-length"] == "134"


def test_get_environ_has_remote_addr():
    result = dict(get_environ({"REMOTE_ADDR": "127.0.0.1"}))
    assert "REMOTE_ADDR" in result
    assert result["REMOTE_ADDR"] == "127.0.0.1"


def test_get_environ_has_server_name():
    result = dict(get_environ({"SERVER_NAME": "127.0.0.1"}))
    assert "SERVER_NAME" in result
    assert result["SERVER_NAME"] == "127.0.0.1"


def test_get_environ_has_server_port():
    result = dict(get_environ({"SERVER_PORT": 80}))
    assert "SERVER_PORT" in result
    assert result["SERVER_PORT"] == 80


def test_get_environ_hides_wsgi_input():
    result = list(get_environ({"wsgi.input": "foo"}))
    assert "wsgi.input" not in result


def test_get_host_http_x_forwarded_host():
    result = get_host({"HTTP_X_FORWARDED_HOST": "example.com"})
    assert result == "example.com"


def test_get_host_http_host():
    result = get_host({"HTTP_HOST": "example.com"})
    assert result == "example.com"


def test_get_host_http_strips_port():
    result = get_host({"wsgi.url_scheme": "http", "SERVER_NAME": "example.com", "SERVER_PORT": "80"})
    assert result == "example.com"


def test_get_host_https_strips_port():
    result = get_host({"wsgi.url_scheme": "https", "SERVER_NAME": "example.com", "SERVER_PORT": "443"})
    assert result == "example.com"


def test_get_host_http_nonstandard_port():
    result = get_host({"wsgi.url_scheme": "http", "SERVER_NAME": "example.com", "SERVER_PORT": "81"})
    assert result == "example.com:81"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
import urllib.parse
from collections import namedtuple

from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule
from elasticapm.traces import capture_span
from elasticapm.utils.logging import get_logger

logger = get_logger("elasticapm.instrument")

HandlerInfo = namedtuple("HandlerInfo", ("signature", "span_type", "span_subtype", "span_action", "context"))


class AzureInstrumentation(AbstractInstrumentedModule):
    name = "azure"

    instrument_list = [
        ("azure.core.pipeline._base", "Pipeline.run"),
        ("azure.cosmosdb.table.common._http.httpclient", "_HTTPClient.perform_request"),
    ]

    def call(self, module, method, wrapped, instance, args, kwargs):
        if len(args) == 1:
            request = args[0]
        else:
            request = kwargs["request"]

        if hasattr(request, "url"):  # Azure Storage HttpRequest
            parsed_url = urllib.parse.urlparse(request.url)
            hostname = parsed_url.hostname
            port = parsed_url.port
            path = parsed_url.path
            query_params = urllib.parse.parse_qs(parsed_url.query)
        else:  # CosmosDB HTTPRequest
            hostname = request.host
            port = hostname.split(":")[1] if ":" in hostname else 80
            path = request.path
            query_params = request.query

        # Detect the service
        service = None
        if ".blob.core." in hostname:
            service = "azureblob"
            service_type = "storage"
        elif ".queue.core." in hostname:
            service = "azurequeue"
            service_type = "messaging"
        elif ".table.core." in hostname:
            service = "azuretable"
            service_type = "storage"
        elif ".file.core." in hostname:
            service = "azurefile"
            service_type = "storage"

        # Do not create a span if we don't recognize the service
        if not service:
            return wrapped(*args, **kwargs)

        context = {
            "destination": {
                "address": hostname,
                "port": port,
            }
        }

        handler_info = handlers[service](request, hostname, path, query_params, service, service_type, context)

        with capture_span(
            handler_info.signature,
            span_type=handler_info.span_type,
            leaf=True,
            span_subtype=handler_info.span_subtype,
            span_action=handler_info.span_action,
            extra=handler_info.context,
        ):
            return wrapped(*args, **kwargs)


def handle_azureblob(request, hostname, path, query_params, service, service_type, context):
    """
    Returns the HandlerInfo for Azure Blob Storage operations
    """
    account_name = hostname.split(".")[0]
    context["destination"]["service"] = {
        "name": service,
        "resource": "{}/{}".format(service, account_name),
        "type": service_type,
    }
    method = request.method
    headers = request.headers
    blob = path[1:]

    operation_name = "Unknown"
    if method.lower() == "delete":
        operation_name = "Delete"
    elif method.lower() == "get":
        operation_name = "Download"
        if "container" in query_params.get("restype", []):
            operation_name = "GetProperties"
            if "acl" in query_params.get("comp", []):
                operation_name = "GetAcl"
            elif "list" in query_params.get("comp", []):
                operation_name = "ListBlobs"
        elif "metadata" in query_params.get("comp", []):
            operation_name = "GetMetadata"
        elif "list" in query_params.get("comp", []):
            operation_name = "ListContainers"
        elif "tags" in query_params.get("comp", []):
            operation_name = "GetTags"
            if query_params.get("where"):
                operation_name = "FindTags"
        elif "blocklist" in query_params.get("comp", []):
            operation_name = "GetBlockList"
        elif "pagelist" in query_params.get("comp", []):
            operation_name = "GetPageRanges"
        elif "stats" in query_params.get("comp", []):
            operation_name = "Stats"
        elif "blobs" in query_params.get("comp", []):
            operation_name = "FilterBlobs"
    elif method.lower() == "head":
        operation_name = "GetProperties"
        if "container" in query_params.get("restype", []) and query_params.get("comp") == "metadata":
            operation_name = "GetMetadata"
        elif "container" in query_params.get("restype", []) and query_params.get("comp") == "acl":
            operation_name = "GetAcl"
    elif method.lower() == "post":
        if "batch" in query_params.get("comp", []):
            operation_name = "Batch"
        elif "query" in query_params.get("comp", []):
            operation_name = "Query"
        elif "userdelegationkey" in query_params.get("comp", []):
            operation_name = "GetUserDelegationKey"
    elif method.lower() == "put":
        operation_name = "Create"
        if "x-ms-copy-source" in headers:
            operation_name = "Copy"
            # These are repetitive and unnecessary, but included in case the table at
            # https://github.com/elastic/apm/blob/main/specs/agents/tracing-instrumentation-azure.md
            # changes in the future
            if "block" in query_params.get("comp", []):
                operation_name = "Copy"
            elif "page" in query_params.get("comp", []):
                operation_name = "Copy"
            elif "incrementalcopy" in query_params.get("comp", []):
                operation_name = "Copy"
            elif "appendblock" in query_params.get("comp", []):
                operation_name = "Copy"
        elif "x-ms-blob-type" in headers:
            operation_name = "Upload"
        elif "x-ms-page-write" in headers and query_params.get("comp") == "page":
            operation_name = "Clear"
        elif "copy" in query_params.get("comp", []):
            operation_name = "Abort"
        elif "block" in query_params.get("comp", []):
            operation_name = "Upload"
        elif "blocklist" in query_params.get("comp", []):
            operation_name = "Upload"
        elif "page" in query_params.get("comp", []):
            operation_name = "Upload"
        elif "appendblock" in query_params.get("comp", []):
            operation_name = "Upload"
        elif "metadata" in query_params.get("comp", []):
            operation_name = "SetMetadata"
        elif "container" in query_params.get("restype", []) and query_params.get("comp") == "acl":
            operation_name = "SetAcl"
        elif "properties" in query_params.get("comp", []):
            operation_name = "SetProperties"
        elif "lease" in query_params.get("comp", []):
            operation_name = "Lease"
        elif "snapshot" in query_params.get("comp", []):
            operation_name = "Snapshot"
        elif "undelete" in query_params.get("comp", []):
            operation_name = "Undelete"
        elif "tags" in query_params.get("comp", []):
            operation_name = "SetTags"
        elif "tier" in query_params.get("comp", []):
            operation_name = "SetTier"
        elif "expiry" in query_params.get("comp", []):
            operation_name = "SetExpiry"
        elif "seal" in query_params.get("comp", []):
            operation_name = "Seal"
        elif "rename" in query_params.get("comp", []):
            operation_name = "Rename"

    signature = "AzureBlob {} {}".format(operation_name, blob)

    return HandlerInfo(signature, service_type, service, operation_name, context)


def handle_azurequeue(request, hostname, path, query_params, service, service_type, context):
    """
    Returns the HandlerInfo for Azure Queue operations
    """
    account_name = hostname.split(".")[0]
    method = request.method
    resource_name = path.split("/")[1] if "/" in path else account_name  # /queuename/messages
    context["destination"]["service"] = {
        "name": service,
        "resource": "{}/{}".format(service, resource_name),
        "type": service_type,
    }

    operation_name = "UNKNOWN"
    preposition = "to "
    if method.lower() == "delete":
        operation_name = "DELETE"
        preposition = ""
        if path.endswith("/messages") and "popreceipt" not in query_params:
            operation_name = "CLEAR"
        elif query_params.get("popreceipt", []):
            # Redundant, but included in case the table at
            # https://github.com/elastic/apm/blob/main/specs/agents/tracing-instrumentation-azure.md
            # changes in the future
            operation_name = "DELETE"
            preposition = "from "
    elif method.lower() == "get":
        operation_name = "RECEIVE"
        preposition = "from "
        if "list" in query_params.get("comp", []):
            operation_name = "LISTQUEUES"
        elif "properties" in query_params.get("comp", []):
            operation_name = "GETPROPERTIES"
        elif "stats" in query_params.get("comp", []):
            operation_name = "STATS"
        elif "metadata" in query_params.get("comp", []):
            operation_name = "GETMETADATA"
        elif "acl" in query_params.get("comp", []):
            operation_name = "GETACL"
        elif "true" in query_params.get("peekonly", []):
            operation_name = "PEEK"
    elif method.lower() == "head":
        operation_name = "RECEIVE"
        preposition = "from "
        if "metadata" in query_params.get("comp", []):
            operation_name = "GETMETADATA"
        elif "acl" in query_params.get("comp", []):
            operation_name = "GETACL"
    elif method.lower() == "options":
        operation_name = "PREFLIGHT"
        preposition = "from "
    elif method.lower() == "post":
        operation_name = "SEND"
        preposition = "to "
    elif method.lower() == "put":
        operation_name = "CREATE"
        preposition = ""
        if "metadata" in query_params.get("comp", []):
            operation_name = "SETMETADATA"
            preposition = "for "
        elif "acl" in query_params.get("comp", []):
            operation_name = "SETACL"
            preposition = "for "
        elif "properties" in query_params.get("comp", []):
            operation_name = "SETPROPERTIES"
            preposition = "for "
        elif query_params.get("popreceipt", []):
            operation_name = "UPDATE"
            preposition = ""

    # If `preposition` is included, it should have a trailing space
    signature = "AzureQueue {} {}{}".format(operation_name, preposition, resource_name)

    return HandlerInfo(signature, service_type, service, operation_name.lower(), context)


def handle_azuretable(request, hostname, path, query_params, service, service_type, context):
    """
    Returns the HandlerInfo for Azure Table Storage operations
    """
    account_name = hostname.split(".")[0]
    method = request.method
    body = request.body
    if body:
        try:
            body = json.loads(body)
        except json.decoder.JSONDecodeError:  # str not bytes
            body = {}
    else:
        body = {}
    # /tablename(PartitionKey='<partition-key>',RowKey='<row-key>')
    resource_name = path.split("/", 1)[1] if "/" in path else path
    context["destination"]["service"] = {
        "name": service,
        "resource": "{}/{}".format(service, account_name),
        "type": service_type,
    }

    operation_name = "Unknown"
    if method.lower() == "put" or method.lower() == "patch":
        operation_name = "Update"
        if "properties" in query_params.get("comp", []):
            operation_name = "SetProperties"
        elif "acl" in query_params.get("comp", []):
            operation_name = "SetAcl"
    elif method.lower() == "post":
        if resource_name == "Tables":
            resource_name = body.get("TableName", resource_name)
            operation_name = "Create"
        else:  # /<tablename>
            operation_name = "Insert"
    elif method.lower() == "get":
        operation_name = "Query"  # for both /Tables and /table()
        if "properties" in query_params.get("comp", []):
            operation_name = "GetProperties"
        elif "stats" in query_params.get("comp", []):
            operation_name = "Stats"
        elif "acl" in query_params.get("comp", []):
            operation_name = "GetAcl"
    elif method.lower() == "delete":
        operation_name = "Delete"
        if "Tables" in resource_name and "'" in resource_name:
            resource_name = resource_name.split("'")[1]  # /Tables('<table_name>')
    elif method.lower() == "options":
        operation_name = "Preflight"
    elif method.lower() == "head" and "acl" in query_params.get("comp", []):
        operation_name = "GetAcl"
    elif method.lower() == "merge":
        operation_name = "Merge"

    signature = "AzureTable {} {}".format(operation_name, resource_name)

    return HandlerInfo(signature, service_type, service, operation_name, context)


def handle_azurefile(request, hostname, path, query_params, service, service_type, context):
    """
    Returns the HandlerInfo for Azure File Share Storage operations
    """
    account_name = hostname.split(".")[0]
    method = request.method
    resource_name = path.split("/", 1)[1] if "/" in path else account_name
    headers = request.headers
    context["destination"]["service"] = {
        "name": service,
        "resource": "{}/{}".format(service, account_name),
        "type": service_type,
    }

    operation_name = "Unknown"
    if method.lower() == "get":
        operation_name = "Download"
        if "list" in query_params.get("comp", []):
            operation_name = "List"
        elif "properties" in query_params.get("comp", []):
            operation_name = "GetProperties"
        elif "share" in query_params.get("restype", []):
            operation_name = "GetProperties"
        elif "metadata" in query_params.get("comp", []):
            operation_name = "GetMetadata"
        elif "acl" in query_params.get("comp", []):
            operation_name = "GetAcl"
        elif "stats" in query_params.get("comp", []):
            operation_name = "Stats"
        elif "filepermission" in query_params.get("comp", []):
            operation_name = "GetPermission"
        elif "listhandles" in query_params.get("comp", []):
            operation_name = "ListHandles"
        elif "rangelist" in query_params.get("comp", []):
            operation_name = "ListRanges"
    elif method.lower() == "put":
        operation_name = "Create"
        if "properties" in query_params.get("comp", []):
            operation_name = "SetProperties"
            if "share" in query_params.get("restype", []):
                operation_name = "SetProperties"
        elif "snapshot" in query_params.get("comp", []):
            operation_name = "Snapshot"
        elif "metadata" in query_params.get("comp", []):
            operation_name = "SetMetadata"
        elif "undelete" in query_params.get("comp", []):
            operation_name = "Undelete"
        elif "acl" in query_params.get("comp", []):
            operation_name = "SetAcl"
        elif "filepermission" in query_params.get("comp", []):
            operation_name = "SetPermission"
        elif "directory" in query_params.get("restype", []):
            operation_name = "Create"
        elif "forceclosehandles" in query_params.get("comp", []):
            operation_name = "CloseHandles"
        elif "range" in query_params.get("comp", []):
            operation_name = "Upload"
        elif "x-ms-copy-source" in headers:
            operation_name = "Copy"
        elif "x-ms-copy-action" in headers and headers["x-ms-copy-action"] == "abort":
            operation_name = "Abort"
        elif "lease" in query_params.get("comp", []):
            operation_name = "Lease"
    elif method.lower() == "options":
        operation_name = "Preflight"
    elif method.lower() == "head":
        operation_name = "GetProperties"
        if "share" in query_params.get("restype", []):
            operation_name = "GetProperties"
        elif "metadata" in query_params.get("comp", []):
            operation_name = "GetMetadata"
        elif "acl" in query_params.get("comp", []):
            operation_name = "GetAcl"
    elif method.lower() == "delete":
        operation_name = "Delete"

    signature = "AzureFile {} {}".format(operation_name, resource_name)

    return HandlerInfo(signature, service_type, service, operation_name, context)


handlers = {
    "azureblob": handle_azureblob,
    "azurequeue": handle_azurequeue,
    "azuretable": handle_azuretable,
    "azurefile": handle_azurefile,
}



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import itertools
import os

import pytest
from pytest_bdd import given, parsers, scenario, scenarios, then, when

# TODO fix up these tests with the updated definitions in api_key.feature
# Skipping until then
pytestmark = [pytest.mark.bdd, pytest.mark.skip]

version_counter = itertools.count(0)

scenarios(os.path.join("features", "api_key.feature"))


@given("an agent", target_fixture="an_agent")
def an_agent(elasticapm_client):
    return elasticapm_client


@when(parsers.parse("an api key is set to '{key}' in the config"))
def set_api_key_to_value(an_agent, key):
    an_agent.config.update(version=next(version_counter), api_key=key)


@when("an api key is set in the config")
def set_any_api_key(an_agent):
    an_agent.config.update(next(version_counter), api_key="foo")


@when("a secret_token is set in the config")
def set_any_secret_token(an_agent):
    an_agent.config.update(next(version_counter), secret_token="foo")


@when(parsers.parse("a secret_token is set to '{key}' in the config"))
def set_any_secret_token(an_agent, key):
    an_agent.config.update(next(version_counter), secret_token=key)


@when("an api key is not set in the config")
def unset_api_key(an_agent):
    an_agent.config.update(next(version_counter), api_key=None)


@then(parsers.parse("the Authorization header is '{full_key}'"))
def authorization_full_key(an_agent, key, full_key):
    auth_headers = an_agent._transport.auth_headers
    assert "Authorization" in auth_headers
    assert auth_headers["Authorization"] == full_key


@then("the api key is sent in the Authorization header")
def authorization_api_key(an_agent):
    auth_headers = an_agent._transport.auth_headers
    assert "Authorization" in auth_headers
    assert auth_headers["Authorization"].startswith("ApiKey ")


@then("the secret token is sent in the Authorization header")
def authorization_secret_token(an_agent):
    auth_headers = an_agent._transport.auth_headers
    assert "Authorization" in auth_headers
    assert auth_headers["Authorization"].startswith("Bearer ")



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import logging

from elasticapm.instrumentation.packages.base import AbstractInstrumentedModule

logger = logging.getLogger("elasticapm")


class DjangoAutoInstrumentation(AbstractInstrumentedModule):
    name = "django"

    instrument_list = [("django.apps.registry", "Apps.populate")]

    creates_transactions = True

    def call(self, module, method, wrapped, instance, args, kwargs) -> None:
        installed_apps = list(args[0] if args else kwargs["installed_apps"])
        if (
            "elasticapm.contrib.django" not in installed_apps
            and "elasticapm.contrib.django.apps.ElasticAPMConfig" not in installed_apps
        ):
            logger.info("Prepending elasticapm to Django INSTALLED_APPS")
            installed_apps.insert(0, "elasticapm.contrib.django")
            if args:
                args = list(args)
                args[0] = installed_apps
            else:
                kwargs["installed_apps"] = installed_apps
        wrapped(*args, **kwargs)



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logbook
import pytest
from logbook import LogRecord

from elasticapm.conf.constants import ERROR
from elasticapm.handlers.logbook import LogbookHandler


@pytest.fixture()
def logbook_logger():
    return logbook.Logger(__name__)


@pytest.fixture()
def logbook_handler(elasticapm_client):
    elasticapm_client.config.include_paths = ["tests", "elasticapm"]
    return LogbookHandler(elasticapm_client)


def test_logbook_logger_error_level(logbook_logger, logbook_handler):
    with logbook_handler.applicationbound():
        logbook_logger.error("This is a test error")

    assert len(logbook_handler.client.events) == 1
    event = logbook_handler.client.events[ERROR][0]
    assert event["log"]["logger_name"] == __name__
    assert event["log"]["level"] == "error"
    assert event["log"]["message"] == "This is a test error"
    assert "stacktrace" in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test error"


def test_logger_warning_level(logbook_logger, logbook_handler):
    with logbook_handler.applicationbound():
        logbook_logger.warning("This is a test warning")
    assert len(logbook_handler.client.events) == 1
    event = logbook_handler.client.events[ERROR][0]
    assert event["log"]["logger_name"] == __name__
    assert event["log"]["level"] == "warning"
    assert event["log"]["message"] == "This is a test warning"
    assert "stacktrace" in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test warning"


def test_logger_without_stacktrace_config(logbook_logger, logbook_handler):
    logbook_handler.client.config.auto_log_stacks = False

    with logbook_handler.applicationbound():
        logbook_logger.warning("This is a test warning")

    event = logbook_handler.client.events[ERROR][0]
    assert "stacktrace" not in event["log"]


def test_logger_without_stacktrace_stack_false(logbook_logger, logbook_handler):
    logbook_handler.client.config.auto_log_stacks = True

    with logbook_handler.applicationbound():
        logbook_logger.warning("This is a test warning", stack=False)

    event = logbook_handler.client.events[ERROR][0]
    assert "stacktrace" not in event["log"]


def test_logger_with_extra(logbook_logger, logbook_handler):
    with logbook_handler.applicationbound():
        logbook_logger.info("This is a test info with a url", extra=dict(url="http://example.com"))
    assert len(logbook_handler.client.events) == 1
    event = logbook_handler.client.events[ERROR][0]
    assert event["context"]["custom"]["url"] == "http://example.com"
    assert "stacktrace" in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test info with a url"


def test_logger_with_exc_info(logbook_logger, logbook_handler):
    with logbook_handler.applicationbound():
        try:
            raise ValueError("This is a test ValueError")
        except ValueError:
            logbook_logger.info("This is a test info with an exception", exc_info=True)

    assert len(logbook_handler.client.events) == 1
    event = logbook_handler.client.events[ERROR][0]

    assert event["log"]["message"] == "This is a test info with an exception"
    assert "stacktrace" in event["log"]
    assert "exception" in event
    exc = event["exception"]
    assert exc["type"] == "ValueError"
    assert exc["message"] == "ValueError: This is a test ValueError"
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test info with an exception"


def test_logger_param_message(logbook_logger, logbook_handler):
    with logbook_handler.applicationbound():
        logbook_logger.info("This is a test of %s", "args")
    assert len(logbook_handler.client.events) == 1
    event = logbook_handler.client.events[ERROR][0]
    assert event["log"]["message"] == "This is a test of args"
    assert "stacktrace" in event["log"]
    assert "exception" not in event
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "This is a test of %s"


def test_client_arg(elasticapm_client):
    handler = LogbookHandler(elasticapm_client)
    assert handler.client == elasticapm_client


def test_client_kwarg(elasticapm_client):
    handler = LogbookHandler(client=elasticapm_client)
    assert handler.client == elasticapm_client


def test_invalid_first_arg_type():
    with pytest.raises(ValueError):
        LogbookHandler(object)


def test_missing_client_arg():
    with pytest.raises(TypeError):
        LogbookHandler()


def test_logbook_handler_emit_error(capsys, elasticapm_client):
    handler = LogbookHandler(elasticapm_client)
    handler._emit = lambda: 1 / 0
    handler.emit(LogRecord("x", 1, "Oops"))
    out, err = capsys.readouterr()
    assert "Top level ElasticAPM exception caught" in err
    assert "Oops" in err


def test_logbook_handler_emit_error_non_str_message(capsys, elasticapm_client):
    handler = LogbookHandler(elasticapm_client)
    handler._emit = lambda record: 1 / 0
    handler.emit(LogRecord("x", 1, ValueError("oh no")))
    out, err = capsys.readouterr()
    assert "Top level ElasticAPM exception caught" in err
    assert "oh no" in err


def test_logbook_handler_dont_emit_elasticapm(capsys, elasticapm_client):
    handler = LogbookHandler(elasticapm_client)
    handler.emit(LogRecord("elasticapm.errors", 1, "Oops"))
    out, err = capsys.readouterr()
    assert "Oops" in err


def test_arbitrary_object(elasticapm_client, logbook_logger, logbook_handler):
    with logbook_handler.applicationbound():
        logbook_logger.info(["a", "list", "of", "strings"])
    assert len(logbook_handler.client.events) == 1
    event = logbook_handler.client.events[ERROR][0]
    assert "param_message" in event["log"]
    assert event["log"]["param_message"] == "['a', 'list', 'of', 'strings']"



# -*- coding: utf-8 -*-

#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import importlib
import os

import pytest
from mock import Mock

import elasticapm
from elasticapm.conf import constants
from elasticapm.utils import stacks
from elasticapm.utils.stacks import get_culprit, get_stack_info
from tests.utils.stacks import get_me_a_test_frame, get_me_more_test_frames


def nested_frames(count=10):
    if count == 0:
        return


class Context(object):
    def __init__(self, dict):
        self.dict = dict

    __getitem__ = lambda s, *a: s.dict.__getitem__(*a)
    __setitem__ = lambda s, *a: s.dict.__setitem__(*a)
    iterkeys = lambda s, *a: s.dict.keys(*a)


def test_get_culprit_bad_module():
    culprit = get_culprit([{"module": None, "function": "foo"}])
    assert culprit == "<unknown>.foo"

    culprit = get_culprit([{"module": "foo", "function": None}])
    assert culprit == "foo.<unknown>"

    culprit = get_culprit([{}])
    assert culprit == "<unknown>.<unknown>"


def test_bad_locals_in_frame():
    frame = Mock()
    frame.f_locals = Context({"foo": "bar", "biz": "baz"})
    frame.f_lineno = 1
    frame.f_globals = {}
    frame.f_code.co_filename = __file__.replace(".pyc", ".py")
    frame.f_code.co_name = __name__
    frames = [(frame, 1)]
    results = get_stack_info(frames)
    assert len(results) == 1
    result = results[0]
    assert "vars" in result
    variables = {"foo": "bar", "biz": "baz"}
    assert result["vars"] == variables


def test_traceback_hide(elasticapm_client):
    def get_me_a_filtered_frame(hide=True):
        __traceback_hide__ = True
        if not hide:
            del __traceback_hide__

        return list(stacks.iter_stack_frames())

    # hide frame from `get_me_a_filtered_frame
    frames = list(stacks.get_stack_info(get_me_a_filtered_frame(hide=True)))
    assert frames[0]["function"] == "test_traceback_hide"

    # don't hide it:
    frames = list(stacks.get_stack_info(get_me_a_filtered_frame(hide=False)))
    assert frames[0]["function"] == "get_me_a_filtered_frame"


def test_iter_stack_frames_skip_frames():
    frames = get_me_more_test_frames(4)

    iterated_frames = list(stacks.iter_stack_frames(frames, skip=3))
    assert len(iterated_frames) == 1
    assert iterated_frames[0][0].f_locals["count"] == 4


def test_iter_stack_frames_skip_frames_by_module():
    frames = [
        Mock(f_lineno=1, f_globals={"__name__": "foo.bar"}),
        Mock(f_lineno=2, f_globals={"__name__": "foo.bar"}),
        Mock(f_lineno=3, f_globals={"__name__": "baz.bar"}),
        Mock(f_lineno=4, f_globals={"__name__": "foo.bar"}),
    ]

    iterated_frames = list(stacks.iter_stack_frames(frames, skip_top_modules=("foo.",)))
    assert len(iterated_frames) == 2
    assert iterated_frames[0][1] == 3
    assert iterated_frames[1][1] == 4


def test_iter_stack_frames_max_frames():
    iterated_frames = list(stacks.iter_stack_frames(get_me_more_test_frames(10), config=Mock(stack_trace_limit=5)))
    assert len(iterated_frames) == 5
    assert iterated_frames[4][0].f_locals["count"] == 5

    iterated_frames = list(stacks.iter_stack_frames(get_me_more_test_frames(10), config=Mock(stack_trace_limit=-1)))
    assert len(iterated_frames) == 10

    iterated_frames = list(stacks.iter_stack_frames(get_me_more_test_frames(10), config=Mock(stack_trace_limit=0)))
    assert len(iterated_frames) == 0

    iterated_frames = list(
        stacks.iter_stack_frames(get_me_more_test_frames(10), skip=3, config=Mock(stack_trace_limit=5))
    )
    assert len(iterated_frames) == 5
    assert iterated_frames[4][0].f_locals["count"] == 8


@pytest.mark.parametrize(
    "elasticapm_client", [{"stack_trace_limit": 10, "span_stack_trace_min_duration": 0}], indirect=True
)
def test_iter_stack_frames_max_frames_is_dynamic(elasticapm_client):
    def func():
        with elasticapm.capture_span("yay"):
            pass

    elasticapm_client.begin_transaction("foo")
    get_me_more_test_frames(15, func=func)
    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert len(span["stacktrace"]) == 10

    elasticapm_client.config.update(version="2", stack_trace_limit=5)

    elasticapm_client.begin_transaction("foo")
    get_me_more_test_frames(15, func=func)
    elasticapm_client.end_transaction()
    transaction = elasticapm_client.events[constants.TRANSACTION][1]

    span = elasticapm_client.spans_for_transaction(transaction)[0]
    assert len(span["stacktrace"]) == 5


@pytest.mark.parametrize(
    "elasticapm_client", [{"include_paths": ("/a/b/c/*", "/c/d/*"), "exclude_paths": ("/c/*",)}], indirect=True
)
def test_library_frames(elasticapm_client):
    include = elasticapm_client.include_paths_re
    exclude = elasticapm_client.exclude_paths_re
    frame1 = Mock(f_code=Mock(co_filename="/a/b/c/d.py"))
    frame2 = Mock(f_code=Mock(co_filename="/a/b/c/d/e.py"))
    frame3 = Mock(f_code=Mock(co_filename="/c/d/e.py"))
    frame4 = Mock(f_code=Mock(co_filename="/c/e.py"))

    info1 = stacks.get_frame_info(frame1, 1, False, False, include_paths_re=include, exclude_paths_re=exclude)
    info2 = stacks.get_frame_info(frame2, 1, False, False, include_paths_re=include, exclude_paths_re=exclude)
    info3 = stacks.get_frame_info(frame3, 1, False, False, include_paths_re=include, exclude_paths_re=exclude)
    info4 = stacks.get_frame_info(frame4, 1, False, False, include_paths_re=include, exclude_paths_re=exclude)
    assert not info1["library_frame"]
    assert not info2["library_frame"]
    assert not info3["library_frame"]
    assert info4["library_frame"]


def test_get_frame_info():
    frame = get_me_a_test_frame()
    frame_info = stacks.get_frame_info(
        frame, frame.f_lineno, library_frame_context_lines=5, in_app_frame_context_lines=5, with_locals=True
    )
    assert frame_info["function"] == "get_me_a_test_frame"
    assert frame_info["filename"] == os.path.join("tests", "utils", "stacks", "__init__.py")
    assert frame_info["module"] == "tests.utils.stacks"
    assert frame_info["lineno"] == 36
    assert frame_info["context_metadata"][0].endswith(frame_info["filename"])
    assert frame_info["context_metadata"][1] == frame_info["lineno"]
    assert frame_info["context_metadata"][4] == frame_info["module"]
    assert frame_info["vars"] == {"a_local_var": 42}


@pytest.mark.parametrize(
    "lineno,context,expected",
    [
        (10, 5, (["5", "6", "7", "8", "9"], "10", ["11", "12", "13", "14", "15"])),
        (1, 5, ([], "1", ["2", "3", "4", "5", "6"])),
        (2, 5, (["1"], "2", ["3", "4", "5", "6", "7"])),
        (20, 5, (["15", "16", "17", "18", "19"], "20", [])),
        (19, 5, (["14", "15", "16", "17", "18"], "19", ["20"])),
        (1, 0, ([], "1", [])),
        (21, 0, (None, None, None)),
    ],
)
def test_get_lines_from_file(lineno, context, expected):
    stacks.get_lines_from_file.cache_clear()
    fname = os.path.join(os.path.dirname(__file__), "linenos.py")
    result = stacks.get_lines_from_file(fname, lineno, context)
    assert result == expected


@pytest.mark.parametrize(
    "lineno,context,expected",
    [
        (10, 5, (["5", "6", "7", "8", "9"], "10", ["11", "12", "13", "14", "15"])),
        (1, 5, ([], "1", ["2", "3", "4", "5", "6"])),
        (2, 5, (["1"], "2", ["3", "4", "5", "6", "7"])),
        (20, 5, (["15", "16", "17", "18", "19"], "20", [])),
        (19, 5, (["14", "15", "16", "17", "18"], "19", ["20"])),
        (1, 0, ([], "1", [])),
        (21, 0, (None, None, None)),
    ],
)
def test_get_lines_from_loader(lineno, context, expected):
    stacks.get_lines_from_file.cache_clear()
    module = "tests.utils.stacks.linenos"
    spec = importlib.util.find_spec(module)
    loader = spec.loader if spec is not None else None
    fname = os.path.join(os.path.dirname(__file__), "linenos.py")
    result = stacks.get_lines_from_file(fname, lineno, context, loader=loader, module_name=module)
    assert result == expected



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("opentelemetry.sdk")  # isort:skip

from opentelemetry.trace import Link, SpanContext, SpanKind, TraceFlags
from opentelemetry.trace.propagation import _SPAN_KEY
from opentelemetry.trace.status import Status, StatusCode

import elasticapm.contrib.opentelemetry.context as context
import elasticapm.contrib.opentelemetry.trace as trace
from elasticapm.conf import constants
from elasticapm.contrib.opentelemetry.trace import Tracer

pytestmark = pytest.mark.opentelemetry


@pytest.fixture
def tracer(elasticapm_client) -> Tracer:
    yield Tracer("test", elasticapm_client=elasticapm_client)


def test_root_transaction(tracer: Tracer):
    with tracer.start_as_current_span("test"):
        pass

    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    assert transaction["type"] == "unknown"
    assert transaction["name"] == "test"
    assert transaction["result"] == "OK"


def test_ot_span(tracer: Tracer):
    with tracer.start_as_current_span("test"):
        with tracer.start_as_current_span("testspan", kind=SpanKind.CONSUMER):
            with tracer.start_as_current_span("testspan2"):
                pass
    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    span1 = client.events[constants.SPAN][1]
    span2 = client.events[constants.SPAN][0]
    assert span1["transaction_id"] == span1["parent_id"] == transaction["id"]
    assert span1["name"] == "testspan"

    assert span2["transaction_id"] == transaction["id"]
    assert span2["parent_id"] == span1["id"]
    assert span2["name"] == "testspan2"


def test_ot_span_without_auto_attach(tracer: Tracer):
    with tracer.start_as_current_span("test"):
        span = tracer.start_span("testspan", kind=SpanKind.CONSUMER)
        with tracer.start_as_current_span("testspan2"):
            pass
        with trace.use_span(span, end_on_exit=True):
            with tracer.start_as_current_span("testspan3"):
                pass
    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    # The spans come in out of the normal/intuitive order, since "testspan2" ends first.
    span1 = client.events[constants.SPAN][2]
    span2 = client.events[constants.SPAN][0]
    span3 = client.events[constants.SPAN][1]
    assert span1["transaction_id"] == span1["parent_id"] == transaction["id"]
    assert span1["name"] == "testspan"

    assert span2["transaction_id"] == span2["parent_id"] == transaction["id"]
    assert span2["name"] == "testspan2"

    assert span3["transaction_id"] == transaction["id"]
    assert span3["parent_id"] == span1["id"]
    assert span3["name"] == "testspan3"


def test_ot_transaction_types(tracer: Tracer):
    with tracer.start_as_current_span(
        "test_request", attributes={"http.url": "http://localhost"}, kind=SpanKind.SERVER
    ):
        pass
    with tracer.start_as_current_span(
        "test_messaging", attributes={"messaging.system": "kafka"}, kind=SpanKind.CONSUMER
    ):
        pass
    client = tracer.client
    transaction1 = client.events[constants.TRANSACTION][0]
    transaction2 = client.events[constants.TRANSACTION][1]
    assert transaction1["type"] == "request"
    assert transaction2["type"] == "messaging"

    assert transaction1["otel"]["span_kind"] == "SERVER"
    assert transaction1["otel"]["attributes"]["http.url"] == "http://localhost"


def test_ot_exception(tracer: Tracer):
    with pytest.raises(Exception):
        with tracer.start_as_current_span("test"):
            raise Exception()

    client = tracer.client
    error = client.events[constants.ERROR][0]
    assert error["context"]["otel_spankind"] == "INTERNAL"


def test_ot_spancontext(tracer: Tracer):
    with tracer.start_as_current_span("test") as ot_span:
        span_context = ot_span.get_span_context()
    assert isinstance(span_context, SpanContext)
    assert span_context.trace_id
    assert span_context.trace_flags.sampled == TraceFlags.SAMPLED


def test_span_links(tracer: Tracer):
    span_context = SpanContext(
        trace_id=int("aabbccddeeff00112233445566778899", 16), span_id=int("0011223344556677", 16), is_remote=False
    )
    link = Link(span_context)
    with tracer.start_as_current_span("testtransaction", links=[link]):
        with tracer.start_as_current_span("testspan", links=[link]):
            pass
    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    span = client.events[constants.SPAN][0]
    assert transaction["links"][0]["trace_id"] == "aabbccddeeff00112233445566778899"
    assert transaction["links"][0]["span_id"] == "0011223344556677"
    assert span["links"][0]["trace_id"] == "aabbccddeeff00112233445566778899"
    assert span["links"][0]["span_id"] == "0011223344556677"


def test_set_status_with_status_object(tracer: Tracer):
    with tracer.start_as_current_span("test") as span:
        span.set_status(Status(StatusCode.OK))

    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    assert transaction["outcome"] == "success"


def test_set_status_with_status_code(tracer: Tracer):
    with tracer.start_as_current_span("test") as span:
        span.set_status(StatusCode.ERROR)

    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    assert transaction["outcome"] == "failure"


def test_set_status_with_status_code_and_description(tracer: Tracer):
    with tracer.start_as_current_span("test") as span:
        span.set_status(StatusCode.OK, "Everything is fine")

    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    assert transaction["outcome"] == "success"


def test_set_status_unset(tracer: Tracer):
    with tracer.start_as_current_span("test") as span:
        span.set_status(StatusCode.UNSET)

    client = tracer.client
    transaction = client.events[constants.TRANSACTION][0]
    assert transaction["outcome"] == "unknown"


def test_set_status_on_span(tracer: Tracer):
    """Test set_status on a child span (not transaction)"""
    with tracer.start_as_current_span("test"):
        with tracer.start_as_current_span("testspan") as span:
            span.set_status(StatusCode.ERROR)

    client = tracer.client
    span_event = client.events[constants.SPAN][0]
    assert span_event["outcome"] == "failure"


# TODO Add some span subtype testing?



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

pytest.importorskip("jinja2")  # isort:skip

import os

from jinja2 import Environment, FileSystemLoader
from jinja2.environment import Template

from elasticapm.conf.constants import TRANSACTION

pytestmark = [pytest.mark.jinja2]


@pytest.fixture()
def jinja_env():
    filedir = os.path.dirname(__file__)
    loader = FileSystemLoader(filedir)
    return Environment(loader=loader)


def test_from_file(instrument, jinja_env, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    template = jinja_env.get_template("mytemplate.html")
    template.render()
    elasticapm_client.end_transaction("MyView")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    expected_signatures = {"mytemplate.html"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "mytemplate.html"
    assert spans[0]["type"] == "template"
    assert spans[0]["subtype"] == "jinja2"
    assert spans[0]["action"] == "render"


def test_from_string(instrument, elasticapm_client):
    elasticapm_client.begin_transaction("transaction.test")
    template = Template("<html></html")
    template.render()
    elasticapm_client.end_transaction("test")

    transactions = elasticapm_client.events[TRANSACTION]
    spans = elasticapm_client.spans_for_transaction(transactions[0])

    expected_signatures = {"<template>"}

    assert {t["name"] for t in spans} == expected_signatures

    assert spans[0]["name"] == "<template>"
    assert spans[0]["type"] == "template"
    assert spans[0]["subtype"] == "jinja2"
    assert spans[0]["action"] == "render"



#  BSD 3-Clause License
#
#  Copyright (c) 2023, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Unfortunately, we can't really do a full end-to-end test for auto-instrumenting Django.
# Installing an app is not reversible, so using this instrumentation "for real" would
# pollute the Django instance used by pytest.

import mock

from elasticapm.instrumentation.packages.django import DjangoAutoInstrumentation


def test_populate_wrapper_args():
    populate_wrapper = DjangoAutoInstrumentation()
    wrapped = mock.Mock()
    populate_wrapper.call(None, None, wrapped, None, [["django.contrib.admin", "django.contrib.auth"]], {})
    assert wrapped.call_count == 1
    args = wrapped.call_args.args
    assert len(args[0]) == 3
    assert args[0][0] == "elasticapm.contrib.django"


def test_populate_wrapper_args_already_set():
    populate_wrapper = DjangoAutoInstrumentation()
    wrapped = mock.Mock()
    populate_wrapper.call(
        None, None, wrapped, None, [["django.contrib.admin", "django.contrib.auth", "elasticapm.contrib.django"]], {}
    )
    assert wrapped.call_count == 1
    args = wrapped.call_args.args
    assert len(args[0]) == 3
    assert args[0][2] == "elasticapm.contrib.django"


def test_populate_wrapper_kwargs():
    populate_wrapper = DjangoAutoInstrumentation()
    wrapped = mock.Mock()
    populate_wrapper.call(
        None, None, wrapped, None, [], {"installed_apps": ["django.contrib.admin", "django.contrib.auth"]}
    )
    assert wrapped.call_count == 1
    kwargs = wrapped.call_args.kwargs
    assert len(kwargs["installed_apps"]) == 3
    assert kwargs["installed_apps"][0] == "elasticapm.contrib.django"


def test_populate_wrapper_kwargs_already_set():
    populate_wrapper = DjangoAutoInstrumentation()
    wrapped = mock.Mock()
    populate_wrapper.call(
        None,
        None,
        wrapped,
        None,
        [],
        {"installed_apps": ["django.contrib.admin", "django.contrib.auth", "elasticapm.contrib.django"]},
    )
    assert wrapped.call_count == 1
    kwargs = wrapped.call_args.kwargs
    assert len(kwargs["installed_apps"]) == 3
    assert kwargs["installed_apps"][2] == "elasticapm.contrib.django"



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import pytest  # isort:skip

aiohttp = pytest.importorskip("aiohttp")  # isort:skip

import mock
from multidict import MultiDict

import elasticapm
from elasticapm import async_capture_span
from elasticapm.conf import constants
from elasticapm.contrib.aiohttp import ElasticAPM
from elasticapm.contrib.aiohttp.middleware import AioHttpTraceParent
from elasticapm.utils.disttracing import TraceParent

pytestmark = [pytest.mark.aiohttp]


@pytest.fixture
def aioeapm(elasticapm_client):
    async def hello(request):
        with async_capture_span("test"):
            pass
        return aiohttp.web.Response(body=b"Hello, world")

    async def boom(request):
        raise aiohttp.web.HTTPInternalServerError(headers={"boom": "boom"})

    async def raise_ok(request):
        raise aiohttp.web.HTTPOk(body=b"a bit odd but sometimes handy")

    async def exception(request):
        raise ValueError("boom")

    app = aiohttp.web.Application()
    app.router.add_route("GET", "/", hello)
    app.router.add_route("GET", "/hello", hello)
    app.router.add_route("GET", "/boom", boom)
    app.router.add_route("GET", "/raise/ok", raise_ok)
    app.router.add_route("GET", "/exception", exception)
    apm = ElasticAPM(app, elasticapm_client)
    yield apm

    elasticapm.uninstrument()


async def test_get(aiohttp_client, aioeapm):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client
    resp = await client.get("/")
    assert resp.status == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 1
    span = spans[0]

    assert transaction["name"] == "GET /"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["type"] == "request"
    assert transaction["span_count"]["started"] == 1
    request = transaction["context"]["request"]
    request["method"] == "GET"
    request["socket"] == {"remote_address": "127.0.0.1"}

    assert span["name"] == "test"


async def test_transaction_ignore_urls(aiohttp_client, aioeapm):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client
    resp = await client.get("/")
    assert resp.status == 200
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    elasticapm_client.config.update(1, transaction_ignore_urls="x")
    resp = await client.get("/")
    assert resp.status == 200
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 2
    elasticapm_client.config.update(1, transaction_ignore_urls="*,x")
    resp = await client.get("/")
    assert resp.status == 200
    # still only two transaction
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 2


@pytest.mark.parametrize(
    "url,exception",
    (("/boom", "HTTPInternalServerError"), ("/exception", "ValueError")),
)
async def test_exception(aiohttp_client, aioeapm, url, exception):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client
    resp = await client.get(url)
    assert resp.status == 500

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 0

    assert transaction["name"] == f"GET {url}"
    assert transaction["result"] == "HTTP 5xx"
    assert transaction["outcome"] == "failure"
    assert transaction["type"] == "request"
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": "127.0.0.1"}
    assert transaction["context"]["response"]["status_code"] == 500

    assert len(elasticapm_client.events[constants.ERROR]) == 1
    error = elasticapm_client.events[constants.ERROR][0]
    assert error["transaction_id"] == transaction["id"]
    assert error["exception"]["type"] == exception
    assert error["context"]["request"] == transaction["context"]["request"]


async def test_http_exception_below_500(aiohttp_client, aioeapm):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client
    resp = await client.get("/raise/ok")
    assert resp.status == 200

    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1
    transaction = elasticapm_client.events[constants.TRANSACTION][0]
    spans = elasticapm_client.spans_for_transaction(transaction)
    assert len(spans) == 0

    assert transaction["name"] == "GET /raise/ok"
    assert transaction["result"] == "HTTP 2xx"
    assert transaction["outcome"] == "success"
    assert transaction["type"] == "request"
    request = transaction["context"]["request"]
    assert request["method"] == "GET"
    assert request["socket"] == {"remote_address": "127.0.0.1"}
    assert transaction["context"]["response"]["status_code"] == 200

    assert len(elasticapm_client.events[constants.ERROR]) == 0


async def test_capture_headers_is_dynamic(aiohttp_client, aioeapm):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client

    elasticapm_client.config.update("1", capture_headers=True)
    await client.get("/boom")

    elasticapm_client.config.update("2", capture_headers=False)
    await client.get("/boom")
    assert elasticapm_client.config.capture_headers is False

    assert "headers" in elasticapm_client.events[constants.TRANSACTION][0]["context"]["request"]
    assert "headers" in elasticapm_client.events[constants.TRANSACTION][0]["context"]["response"]
    assert "headers" in elasticapm_client.events[constants.ERROR][0]["context"]["request"]

    assert "headers" not in elasticapm_client.events[constants.TRANSACTION][1]["context"]["request"]
    assert "headers" not in elasticapm_client.events[constants.TRANSACTION][1]["context"]["response"]
    assert "headers" not in elasticapm_client.events[constants.ERROR][1]["context"]["request"]


async def test_traceparent_handling(aiohttp_client, aioeapm):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client
    with mock.patch(
        "elasticapm.contrib.aiohttp.middleware.TraceParent.from_string", wraps=TraceParent.from_string
    ) as wrapped_from_string:
        resp = await client.get(
            "/boom",
            headers=(
                (constants.TRACEPARENT_HEADER_NAME, "00-0af7651916cd43dd8448eb211c80319c-b7ad6b7169203331-03"),
                (constants.TRACESTATE_HEADER_NAME, "foo=bar,bar=baz"),
                (constants.TRACESTATE_HEADER_NAME, "baz=bazzinga"),
            ),
        )

    transaction = elasticapm_client.events[constants.TRANSACTION][0]

    assert transaction["trace_id"] == "0af7651916cd43dd8448eb211c80319c"
    assert transaction["parent_id"] == "b7ad6b7169203331"
    assert "foo=bar,bar=baz,baz=bazzinga" in wrapped_from_string.call_args[0]


@pytest.mark.parametrize("headers,expected", ((MultiDict((("a", "1"), ("a", "2"))), "1,2"), (MultiDict(), None)))
async def test_aiohttptraceparent_merge(headers, expected):
    result = AioHttpTraceParent.merge_duplicate_headers(headers, "a")
    assert result == expected


async def test_aiohttp_transaction_ignore_urls(aiohttp_client, aioeapm):
    app = aioeapm.app
    client = await aiohttp_client(app)
    elasticapm_client = aioeapm.client
    resp = await client.get("/hello")
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1

    elasticapm_client.config.update(1, transaction_ignore_urls="/*ello,/world")
    resp = await client.get("/hello")
    assert len(elasticapm_client.events[constants.TRANSACTION]) == 1



#  BSD 3-Clause License
#
#  Copyright (c) 2022, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import grpc
import wrapt

import elasticapm
from elasticapm.conf.constants import OUTCOME
from elasticapm.contrib.grpc.utils import STATUS_TO_OUTCOME
from elasticapm.utils.disttracing import TraceParent


# from https://github.com/grpc/grpc/issues/18191
def _wrap_rpc_behavior(handler, continuation):
    if handler is None:
        return None

    if handler.request_streaming and handler.response_streaming:
        behavior_fn = handler.stream_stream
        handler_factory = grpc.stream_stream_rpc_method_handler
    elif handler.request_streaming and not handler.response_streaming:
        behavior_fn = handler.stream_unary
        handler_factory = grpc.stream_unary_rpc_method_handler
    elif not handler.request_streaming and handler.response_streaming:
        behavior_fn = handler.unary_stream
        handler_factory = grpc.unary_stream_rpc_method_handler
    else:
        behavior_fn = handler.unary_unary
        handler_factory = grpc.unary_unary_rpc_method_handler

    return handler_factory(
        continuation(behavior_fn, handler.request_streaming, handler.response_streaming),
        request_deserializer=handler.request_deserializer,
        response_serializer=handler.response_serializer,
    )


def get_trace_parent(handler_call_details):
    traceparent, tracestate = None, None
    for metadata in handler_call_details.invocation_metadata:
        if metadata.key == "traceparent":
            traceparent = metadata.value
        elif metadata.key == "tracestate":
            tracestate = metadata.key
    if traceparent:
        return TraceParent.from_string(traceparent, tracestate)
    else:
        return None


class _ServicerContextWrapper(wrapt.ObjectProxy):
    def __init__(self, wrapped, transaction) -> None:
        self._self_transaction = transaction
        super().__init__(wrapped)

    def abort(self, code, details):
        transaction = self._self_transaction
        if transaction:
            transaction.set_failure()
        return self.__wrapped__.abort(code, details)

    def set_code(self, code):
        transaction = self._self_transaction
        if transaction:
            outcome = STATUS_TO_OUTCOME.get(code, OUTCOME.SUCCESS)
            transaction.set_success() if outcome is OUTCOME.SUCCESS else transaction.set_failure()
        return self.__wrapped__.set_code(code)


class _ServerInterceptor(grpc.ServerInterceptor):
    def intercept_service(self, continuation, handler_call_details):
        def transaction_wrapper(behavior, request_streaming, response_streaming):
            def _interceptor(request_or_iterator, context):
                if request_streaming or response_streaming:  # only unary-unary is supported
                    return behavior(request_or_iterator, context)
                tp = get_trace_parent(handler_call_details)
                client = elasticapm.get_client()
                transaction = client.begin_transaction("request", trace_parent=tp)
                try:
                    result = behavior(request_or_iterator, _ServicerContextWrapper(context, transaction))
                    if transaction and not transaction.outcome:
                        transaction.set_success()
                    return result
                except Exception:
                    if transaction:
                        transaction.set_failure()
                    client.capture_exception(handled=False)
                    raise
                finally:
                    client.end_transaction(name=handler_call_details.method)

            return _interceptor

        return _wrap_rpc_behavior(continuation(handler_call_details), transaction_wrapper)



#  BSD 3-Clause License
#
#  Copyright (c) 2021, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
import elasticapm
from elasticapm.conf.constants import SPAN


def test_config_snapshotting_span_compression_exact_match(elasticapm_client):
    elasticapm_client.config.update(
        version="1",
        **{
            "span_compression_enabled": "true",
            "span_compression_exact_match_max_duration": "50ms",
            "span_compression_same_kind_max_duration": "10ms",
        }
    )
    elasticapm_client.begin_transaction("foo")
    elasticapm_client.config.update(
        version="2", **{"span_compression_enabled": "false", "span_compression_exact_match_max_duration": "20ms"}
    )
    with elasticapm.capture_span(
        "x",
        leaf=True,
        span_type="a",
        span_subtype="b",
        span_action="c",
        extra={"destination": {"service": {"resource": "x"}}},
        duration=0.025,
    ):
        pass
    with elasticapm.capture_span(
        "x",
        leaf=True,
        span_type="a",
        span_subtype="b",
        span_action="c",
        extra={"destination": {"service": {"resource": "x"}}},
        duration=0.025,
    ):
        pass
    elasticapm_client.end_transaction()
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    assert spans[0]["composite"]["compression_strategy"] == "exact_match"


def test_config_snapshotting_span_compression_same_kind(elasticapm_client):
    elasticapm_client.config.update(
        version="1", **{"span_compression_enabled": "true", "span_compression_same_kind_max_duration": "50ms"}
    )
    elasticapm_client.begin_transaction("foo")
    elasticapm_client.config.update(
        version="2", **{"span_compression_enabled": "false", "span_compression_same_kind_max_duration": "20ms"}
    )
    with elasticapm.capture_span(
        "x",
        leaf=True,
        span_type="a",
        span_subtype="b",
        span_action="c",
        extra={"destination": {"service": {"resource": "x"}}},
        duration=0.025,
    ):
        pass
    with elasticapm.capture_span(
        "y",
        leaf=True,
        span_type="a",
        span_subtype="b",
        span_action="c",
        extra={"destination": {"service": {"resource": "x"}}},
        duration=0.025,
    ):
        pass
    elasticapm_client.end_transaction()
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1
    assert spans[0]["composite"]["compression_strategy"] == "same_kind"


def test_config_snapshotting_span_compression_drop_exit_span(elasticapm_client):
    elasticapm_client.config.update(version="1", exit_span_min_duration="10ms")
    elasticapm_client.begin_transaction("foo")
    elasticapm_client.config.update(version="2", exit_span_min_duration="0ms")
    with elasticapm.capture_span(
        "x",
        leaf=True,
        span_type="a",
        span_subtype="b",
        span_action="c",
        extra={"destination": {"service": {"resource": "x"}}},
        duration=0.005,
    ):
        pass
    elasticapm_client.end_transaction()
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 0


def test_config_snapshotting_span_compression_max_spans(elasticapm_client):
    elasticapm_client.config.update(version="1", transaction_max_spans="1")
    elasticapm_client.begin_transaction("foo")
    elasticapm_client.config.update(version="2", transaction_max_spans="100")
    with elasticapm.capture_span("x", leaf=True, span_type="x", span_subtype="y", span_action="z"):
        pass
    with elasticapm.capture_span("x", leaf=True, span_type="a", span_subtype="b", span_action="c"):
        pass
    elasticapm_client.end_transaction()
    spans = elasticapm_client.events[SPAN]
    assert len(spans) == 1



#  BSD 3-Clause License
#
#  Copyright (c) 2019, Elasticsearch BV
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#  * Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
#  * Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
#  * Neither the name of the copyright holder nor the names of its
#    contributors may be used to endorse or promote products derived from
#    this software without specific prior written permission.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
#  DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
#  FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
#  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
#  SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
#  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#  OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
#  OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

import django
from django.conf import settings
from django.http import HttpResponseServerError

from tests.contrib.django.testapp import views

try:
    from django.urls import re_path
except ImportError:
    # Django < 2
    from django.conf.urls import url as re_path


def handler500(request):
    if getattr(settings, "BREAK_THAT_500", False):
        raise ValueError("handler500")
    return HttpResponseServerError("")


urlpatterns = (
    re_path(r"^render-heavy-template$", views.render_template_view, name="render-heavy-template"),
    re_path(r"^render-user-template$", views.render_user_view, name="render-user-template"),
    re_path(r"^no-error$", views.no_error, name="elasticapm-no-error"),
    re_path(r"^no-error-slash/$", views.no_error, name="elasticapm-no-error-slash"),
    re_path(r"^http-error/(?P<status>[0-9]{3})$", views.http_error, name="elasticapm-http-error"),
    re_path(r"^logging$", views.logging_view, name="elasticapm-logging"),
    re_path(r"^ignored-exception/$", views.ignored_exception, name="elasticapm-ignored-exception"),
    re_path(r"^fake-login$", views.fake_login, name="elasticapm-fake-login"),
    re_path(r"^trigger-500$", views.raise_exc, name="elasticapm-raise-exc"),
    re_path(r"^trigger-500-ioerror$", views.raise_ioerror, name="elasticapm-raise-ioerror"),
    re_path(r"^trigger-500-decorated$", views.decorated_raise_exc, name="elasticapm-raise-exc-decor"),
    re_path(r"^trigger-500-django$", views.django_exc, name="elasticapm-django-exc"),
    re_path(r"^trigger-500-django-orm-exc$", views.django_queryset_error, name="elasticapm-django-orm-exc"),
    re_path(r"^trigger-500-template$", views.template_exc, name="elasticapm-template-exc"),
    re_path(r"^trigger-500-log-request$", views.logging_request_exc, name="elasticapm-log-request-exc"),
    re_path(r"^streaming$", views.streaming_view, name="elasticapm-streaming-view"),
    re_path(r"^name-override$", views.override_transaction_name_view, name="elasticapm-name-override"),
    re_path(r"^class-based$", views.ClassBasedView.as_view(), name="elasticapm-class-based"),
)


if django.VERSION >= (1, 8):
    urlpatterns += (re_path(r"^render-jinja2-template$", views.render_jinja2_template, name="render-jinja2-template"),)

if django.VERSION >= (2, 2):
    from django.urls import path

    urlpatterns += (
        path("route/<int:id>/", views.no_error, name="route-view"),
        path("", views.no_error, name="home-view"),
    )
